#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "init.h"
#include "cc2500.h"
#include "rgbtohsv.h"
#include "convtab.h"

#define CS	PB2
#define SI	PB4 //Watch out: this is Serial In on the side of the uC but SO at the CC2500
#define SO	PB3 //Watch out: this is Serial Out on the side of the uC but SI at the CC2500
#define CLK	PB5 

#define ERR_CHIP_NOT_READY	1
#define ERR_CHIP_IS_NO_CC2500	2
#define ERR_CHIP_SETCHAN_FAILED 3
#define ERR_CHIP_INIT_ERROR	4 //readback of a reg failed


volatile unsigned char rgbdata[9];
volatile char update_rgb=3;

//Errors out by sending a code out, morse-wise, on the led. Long=1,
//short=0, lsb first.
void error_out(char code) {
    int x,i;
    char b;
    DDRC|=(1<<LED1);
    while(1) {
    	PORTC&=~(1<<LED1);
	for (x=0; x<100; x++) _delay_ms(10);
	b=1;
	for (i=0; i<8; i++) {
    	    PORTC|=(1<<LED1);
	    for (x=0; x<((code&b)?40:20); x++) _delay_ms(10);
    	    PORTC&=~(1<<LED1);
	    for (x=0; x<((code&b)?20:40); x++) _delay_ms(10);
	    b<<=1;
	}
    }
}


inline void setrgb(unsigned char* dat) {
    int x;
    static int t;
    char u=0;
    for (x=0; x<9; x++) {
	if (rgbdata[x]!=dat[x]) u=1;
    }
    if (u==0) t++;
    if (t>15) u=1;

    if (u) {
	for (x=0; x<9; x++) rgbdata[x]=dat[x];
	update_rgb=3;
	t=0;
    }

}


unsigned char addrs[][10]={
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}};


void writecfg() {
    int x,y;
    unsigned char *p=0;
    eeprom_write_byte(p++,0x55);
    for (y=0; y<3; y++) {
	for (x=0; x<11; x++) {
	    eeprom_busy_wait();
	    eeprom_write_byte(p++,addrs[y][x]);
	}
    }
}

void readcfg() {
    int x,y;
    unsigned char *p=0;
    //Refuse to read uninited eeprom
    if (eeprom_read_byte(p++)!=0x55) return;
    for (y=0; y<3; y++) {
	for (x=0; x<11; x++) {
	    addrs[y][x]=eeprom_read_byte(p++);
	}
    }

}


void learncode(unsigned char* code) {
    //Check if code is known
    int x,y;
    char found,codeissame;
    found=0;
    for (y=0; y<3; y++) {
	codeissame=1;
	if (addrs[y][0]==0) codeissame=0;
	for (x=0; x<10; x++) {
	    if (code[x]!=addrs[y][x+1]) codeissame=0;
	}
	if (codeissame) found=1;
    }
    //Found = 1 means we already got that code. Exit if that's the case.
    if (found) return;

    //Find new slot
    y=0;
    while (addrs[y][0]==1 && y<3) y++;
    if (y==3) return; //full
    
    addrs[y][0]=1;
    for (x=0; x<10; x++) {
	addrs[y][x+1]=code[x];
    }
}

//The LivingColors idea of hue differs a bit from the hue everyone else;
//while the r->y->g->c->b->m->r-idea persists, it's not completely
//linear. The non-linearity is solved by looking the hue to send up in a
//lookuptable defined in convtab.h.
unsigned char adjust_hue_for_livcol(unsigned char realhue) {
    return pgm_read_byte(hue_convtab+realhue);
}

void update_lamps(unsigned char code) {
    update_rgb=3;
    update_a_lamp(code);
}

//Codes: 3-set color, 7-off, 5-on
void update_a_lamp(unsigned char code) {
    static int lamp;
    static char seq;
    int j;
    unsigned char h,s,v;


    if (update_rgb>0) {
    	PORTC|=(1<<LED1);
	if (addrs[lamp][0]!=0) {
	    seq++;
	    rgb_to_hsv(rgbdata[lamp],rgbdata[lamp+3],rgbdata[lamp+6],&h,&s,&v);
	    //wait for a prev packet to be completely sent
	    disablecs();
	    while((ccsendbyte(0xF5)&0x1F)>1) {
	        disablecs();
	    }
	    ccsendbyte(0x36);
	    disablecs();
	    ccsendbyte(0x7F);
	    ccsendbyte(0x0E);
	    for (j=0; j<9; j++) {
	        ccsendbyte(addrs[lamp][j+1]);
	    }
	    ccsendbyte(code);
	    ccsendbyte(seq++);
	    ccsendbyte(adjust_hue_for_livcol(h));
	    ccsendbyte(s);
	    ccsendbyte(v);
	    disablecs();
	    ccsendbyte(0x35);
	    disablecs();
	}
	lamp++; 
	if (lamp>2) lamp=0;

	PORTC&=~(1<<LED1);
	update_rgb--;
    }
}


inline void disablecs(void) {
    PORTB|=(1<<CS);
    _delay_us(150);
    _delay_us(60);
}

unsigned char ccsendbyte(unsigned char n) {
    char t;
    char r;
    if (PORTB&(1<<CS)) {
	PORTB&=~(1<<CS);
	_delay_us(150);
	_delay_us(60);
    }
    r=0;
    for (t=0; t<8; t++) {
	if (n&0x80) PORTB|=(1<<SO); else PORTB&=~(1<<SO);
	PORTB&=~(1<<CLK);
	_delay_us(2);
	PORTB|=(1<<CLK);
	r<<=1;
	if (PINB&(1<<SI)) r|=1;
	_delay_us(2);
	n<<=1;
    }
    PORTB&=~(1<<CLK);
    return r;
}

void setreg(unsigned char r, unsigned char d) {
    unsigned char db;
    disablecs();
    ccsendbyte(r);
    ccsendbyte(d);
    disablecs();
    ccsendbyte(r|0x80);
    db=ccsendbyte(0);
    if (db!=d) error_out(ERR_CHIP_INIT_ERROR);
    disablecs();
}

void initcc2500(void) {
    short i;
    DDRB=(1<<CS)|(1<<SO)|(1<<CLK);
    DDRC|=(1<<LED1)|(1<<LED2)|(1<<LED3)|(1<<LED4);

    disablecs();
    _delay_ms(10);
    ccsendbyte(0x30);
    disablecs();
    _delay_ms(10);
    ccsendbyte(0x30);
    disablecs();
    _delay_ms(10);

    //Check if partnum = 0x80 (which means the device is a cc2500)    
    if (ccsendbyte(0xF0)&0x80) error_out(ERR_CHIP_NOT_READY);
    if (ccsendbyte(0)!=0x80) error_out(ERR_CHIP_IS_NO_CC2500);
    disablecs();

    setreg(0x0B, 0x09);
    setreg(0x0C, 0x00);
    setreg(0x0D, 0x5D);
    setreg(0x0E, 0x93);
    setreg(0x0F, 0xB1);
    setreg(0x10, 0x2D);
    setreg(0x11, 0x3B);
    setreg(0x12, 0x73);
    setreg(0x13, 0x22);
    setreg(0x14, 0xF8);
    setreg(0x0A, 0x03);
    setreg(0x15, 0x00);
    setreg(0x21, 0xB6);
    setreg(0x22, 0x10);
    setreg(0x18, 0x18);
    setreg(0x19, 0x1D);
    setreg(0x1A, 0x1C);
    setreg(0x1B, 0xC7);
    setreg(0x1C, 0x00);
    setreg(0x1D, 0xB2);
    setreg(0x23, 0xEA);
    setreg(0x24, 0x0A);
    setreg(0x25, 0x00);
    setreg(0x26, 0x11);
    setreg(0x29, 0x59);
    setreg(0x2C, 0x88);
    setreg(0x2D, 0x31);
    setreg(0x2E, 0x0B);
    setreg(0x00, 0x06);
    setreg(0x02, 0x01);
    setreg(0x07, 0x04);
    setreg(0x08, 0x45);
    setreg(0x09, 0x00);
    setreg(0x06, 0xFF);
    setreg(0x03, 0x0D);
    setreg(0x3E, 0xA9);
    ccsendbyte(0x36);
    disablecs();
    ccsendbyte(0x36);
    disablecs();
    ccsendbyte(0x39);
    disablecs();
    ccsendbyte(0x36);
    disablecs();

    readcfg();
}


//Called when button is pressed. Should copy strings and addresses from 
//an existing remote controller.
void GoLearnCodes(void) {
    unsigned char code[17];
    char recved=0;
    unsigned char t;
    int x,l;
    disablecs();
    //Clear codes
    for (x=0; x<3; x++) addrs[x][0]=0;

    //Go receive
    while(recved<15) {
	disablecs();
	ccsendbyte(0x36);
	ccsendbyte(0x34);
	ccsendbyte(0x34);
	//wait for packet
	PORTC|=(1<<LED1);
	do {
	    disablecs();
	    ccsendbyte(0xF8);
	} while ((ccsendbyte(0)&1)==0);
	PORTC&=~(1<<LED1);
	//fetch packet
	disablecs();
	ccsendbyte(0xff);
	l=ccsendbyte(0);
	for (x=0; x<l; x++) {
	    t=ccsendbyte(0);
	    if (x<17) code[x]=t;
	}
	disablecs();
	//Check for correct len and turn-off packet
	if (l==0xe && code[9]==7) {
	    learncode(code);
	    recved++;
	}
    }
    disablecs();
    ccsendbyte(0x36);
    ccsendbyte(0x36);
    disablecs();
    ccsendbyte(0x36);
    ccsendbyte(0x36);
    disablecs();
    
    //enable lights
/*
    l=0x80;
    for (l=0; l<255; l++) {
	//set rgb, force update
	for (x=0; x<9; x++) rgbdata[x]=l;
	update_rgb=1;
	//turn found lamps on
	update_lamps(5);
//	_delay_ms(2);
    }
*/

    //write addresses to eeprom
    writecfg();
}
